#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdbool.h>
#include <limits.h>

#define DELKA_BUFFERU 16384
#define MAX_DELKA_JMENA 128
#define SEPARATORY " ;\t\n"

void precti_jmeno_souboru_a_vyprazdni_buffer(int max_delka, char *prompt, char *s, char *s_bez_pripony)
{
  int delka;
  char *p_tecka;
  printf(prompt);
  fgets(s, max_delka, stdin);
  delka = strlen(s) - 1;
  if (s[delka] == '\n')
    s[delka] = '\0';
  else
    while (getchar() != '\n')
      ;
  strcpy(s_bez_pripony, s);
  p_tecka = strrchr(s_bez_pripony, '.');
  if (p_tecka) {
    s_bez_pripony[p_tecka - s_bez_pripony] = '\0';
  }
}

int preved_retezec_na_int(char *jmeno_souboru, int r, int s, char *cislo_string, int *cislo)
{
  char *odpad;
  bool dlouhe_cislo;
  long cislo_long;
  errno = 0;
  cislo_long = strtol(cislo_string, &odpad, 10);
  dlouhe_cislo = cislo_long > INT_MAX || cislo_long < INT_MIN;
  if (errno != 0 || dlouhe_cislo || cislo_string == odpad) {
    printf("V souboru %s je na radku %d ve sloupci %d neplatne cislo %s!\t%s\n", jmeno_souboru, r + 1, s + 1, cislo_string, sys_errlist[errno]);
    return 1;
  } else {
    *cislo = (int)cislo_long;
  }
  if (*odpad != '\0') {
    printf("Zahozeno \"%s\" na radku %d ve sloupci %d!\n", odpad, r + 1, s + 1);
  }
  return 0;
}

int nacti(char *jmeno_souboru, int *r, int *s, int ***x, int **p_i) {
  FILE *fr;
  bool nalezen_radek_s_daty = false;
  char buffer_radku[DELKA_BUFFERU];
  char *cislo_string;
  int pocet_prvku_matice = 0, i, j;
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    return 1;
  }
  while (fgets(buffer_radku, DELKA_BUFFERU, fr) != NULL) {
    cislo_string = strtok(buffer_radku, SEPARATORY);
    if (cislo_string != NULL) {
      (*r)++;
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    return 1;
  }
  if (!(*r)) {
    printf("Soubor %s nema ani jeden radek s daty.\n", jmeno_souboru);
    return 1;
  }
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    return 1;
  }
  while ((fgets(buffer_radku, DELKA_BUFFERU, fr) != NULL) && !nalezen_radek_s_daty) {
    cislo_string = strtok(buffer_radku, SEPARATORY);
    if (cislo_string != NULL) {
      (*s)++;
      nalezen_radek_s_daty = true;
    }
    while ((cislo_string = strtok(NULL, SEPARATORY)) != NULL) {
      (*s)++;
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    return 1;
  }
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    return 1;
  }
  while (fgets(buffer_radku, DELKA_BUFFERU, fr) != NULL) {
    cislo_string = strtok(buffer_radku, SEPARATORY);
    if (cislo_string != NULL) {
      pocet_prvku_matice++;
    }
    while ((cislo_string = strtok(NULL, SEPARATORY)) != NULL) {
      pocet_prvku_matice++;
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    return 1;
  }
  if (pocet_prvku_matice != (*r) * (*s)) {
    printf("V souboru %s neni pocet prvku matice (%d) rovny poctu prvku na prvnim radku s daty (%d) krat pocet radku s daty (%d).\n", jmeno_souboru, pocet_prvku_matice, *s, *r);
    return 1;
  }
  if ((*p_i = (int *) malloc((*r) * (*s) * sizeof(int))) == NULL) {
    printf("Malo pameti!\n");
    return 1;
  }
  if ((*x = (int **) malloc((*r) * sizeof(int *))) == NULL) {
    printf("Malo pameti!\n");
    return 1;
  }
  for (i = 0; i < (*r); i++) {
    (*x)[i] = *p_i + i * (*s);
  }
  if ((fr = fopen(jmeno_souboru, "r")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    return 1;
  }
  i = 0; j = 0;
  while (fgets(buffer_radku, DELKA_BUFFERU, fr) != NULL) {
    cislo_string = strtok(buffer_radku, SEPARATORY);
    if (cislo_string != NULL) {
      if (preved_retezec_na_int(jmeno_souboru, i, j, cislo_string, &(*x)[i][j])) {
        return 1;
      }
      j++;
      while ((cislo_string = strtok(NULL, SEPARATORY)) != NULL) {
        if (preved_retezec_na_int(jmeno_souboru, i, j, cislo_string, &(*x)[i][j])) {
          return 1;
        }
        j++;
      }
      i++; j = 0;
    }
  }
  if (fclose(fr) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    return 1;
  }
  return 0;
}

void zapis(char *jmeno_souboru, int r, int s, int **x) {
  FILE *fw;
  int i, j;
  if ((fw = fopen(jmeno_souboru, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  for (i = 0; i < r; i++) {
    for (j = 0; j < s; j++) {
      fprintf(fw, "%d ", x[i][j]);
    }
    fprintf(fw, "\n");
  }
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

void uvolni_pamet_pro_matici(int ***x, int **p_i) {
  free((void *) *p_i);
  *p_i = NULL;
  free((void *) *x);
  *x = NULL;
}

void uvolni_pamet_pro_vektor(int **x) {
  free((void *) *x);
  *x = NULL;
}

void zacni(int r, int s, int **x, int **spravny_vystup, int **w) {
  const int PRAH = 1;
  int i;
  if ((*spravny_vystup = (int *) malloc(r * sizeof(int))) == NULL) {
    printf("Malo pameti!\n");
    exit(1);
  }
  if ((*w = (int *) calloc(s, sizeof(int))) == NULL) { /* Funkce calloc() vynuluje vahy. */
    printf("Malo pameti!\n");
    exit(1);
  }
  for (i = 0; i < r; i++) {
    (*spravny_vystup)[i] = x[i][s - 1];
    x[i][s - 1] = PRAH;
  }
}

void pracuj(char *jmeno_souboru, int r, int s, int **x, int *spravny_vystup, int *w) {
  FILE *fw;
  int i, j, pruchody = 0, zmeny_vah = 0, suma;
  bool vahy_hotovo;
  do {
    vahy_hotovo = true;
    for (i = 0; i < r; i++) {
      suma = 0;
      for (j = 0; j < s; j++) {
        suma += x[i][j] * w[j];
      }
      if (suma <= 0 && spravny_vystup[i] == 1) {
        for (j = 0; j < s; j++) {
          w[j] += x[i][j];
        }
        zmeny_vah++;
        vahy_hotovo = false;
      } else if (suma > 0 && spravny_vystup[i] == 0) {
          for (j = 0; j < s; j++) {
            w[j] -= x[i][j];
          }
          zmeny_vah++;
          vahy_hotovo = false;
        }
    }
    pruchody++;
  } while (!vahy_hotovo);
  if ((fw = fopen(jmeno_souboru, "w")) == NULL) {
    printf("Soubor %s se nepodarilo otevrit.\n", jmeno_souboru);
    exit(1);
  }
  fprintf(fw, "V %d. pruchodu trenovaci davkou se vahy jiz nezmenily.\n", pruchody);
  fprintf(fw, "Vektor vah byl zmenen %d krat.\n", zmeny_vah);
  fprintf(fw, "Vektor vah je (");
  for (i = 0; i < s; i++) {
    fprintf(fw, "%d ", w[i]);
  }
  fprintf(fw, ")\n");
  if (fclose(fw) == EOF) {
    printf("Soubor %s se nepodarilo uzavrit.\n", jmeno_souboru);
    exit(1);
  }
}

int main()
{
  bool nenalezen_radek_s_daty;
  int radek = 0, sloupek = 0;
  int **x, *p_i, *spravny_vystup, *w;
  char jmeno_souboru[MAX_DELKA_JMENA] = { '\0' };
  char jmeno_souboru_bez_pripony[MAX_DELKA_JMENA] = { '\0' };
  precti_jmeno_souboru_a_vyprazdni_buffer(MAX_DELKA_JMENA, "Zadej nazev souboru s daty: ", jmeno_souboru, jmeno_souboru_bez_pripony);
  nenalezen_radek_s_daty = nacti(jmeno_souboru, &radek, &sloupek, &x, &p_i);
  if (nenalezen_radek_s_daty) {
    return nenalezen_radek_s_daty;
  }
  sprintf(jmeno_souboru, "%s_vystup.txt", jmeno_souboru_bez_pripony);
  zapis(jmeno_souboru, radek, sloupek, x);
  zacni(radek, sloupek, x, &spravny_vystup, &w);
  sprintf(jmeno_souboru, "%s_vysledek_perceptron.txt", jmeno_souboru_bez_pripony);
  pracuj(jmeno_souboru, radek, sloupek, x, spravny_vystup, w);
  uvolni_pamet_pro_matici(&x, &p_i);
  uvolni_pamet_pro_vektor(&spravny_vystup);
  uvolni_pamet_pro_vektor(&w);
  return 0;
}
